/*
 * credits.c - credits scene
 * Copyright (C) 2009  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <math.h>
#include "credits.h"
#include "scene.h"
#include "util.h"
#include "video.h"
#include "audio.h"
#include "font.h"
#include "actor.h"
#include "lang.h"
#include "input.h"


/* private data */
#define CREDITS_BGFILE             "images/squarebg.png"
#define CREDITS_MAXLINES           8
static image_t *bg, *box;
static v2d_t bgpos;
static int quit;
static font_t *title, *text, *back;
static input_t *input;
static double scene_time;
static int line_count;

/* private functions */
static char* format_dev(char *str);




/* public functions */

/*
 * credits_init()
 * Initializes the scene
 */
void credits_init()
{
    int i;
    char line[CREDITS_MAXLINES][256], *p;

    /* credits */
    strcpy(line[0], lang_get("CREDITS_PROGRAMMERS"));
    sprintf(line[1], "Alexandre Martins\n%s", format_dev(lang_get("CREDITS_DEV_ALEXANDRE")));
    strcpy(line[2], lang_get("CREDITS_MUSICIANS"));
    sprintf(line[3], "Di Rodrigues\n%s", format_dev(lang_get("CREDITS_DEV_DIRODRIGUES")));
    strcpy(line[4], lang_get("CREDITS_CONTRIBUTORS"));
    sprintf(line[5], "Lainz\n%s", format_dev(lang_get("CREDITS_DEV_LAINZ")));
    sprintf(line[6], "Arthur Blot\n%s", format_dev(lang_get("CREDITS_DEV_ARTHURBLOT")));
    sprintf(line[7], "Neoblast\n%s", format_dev(lang_get("CREDITS_DEV_NEOBLAST")));

    /* avoid hacking */
    for(i=0; i<CREDITS_MAXLINES; i++) {
        for(p=line[i]; *p; p++) {
            if(*p == '\\' && *(p+1) == 'n')
                *p = *(p+1) = ' ';
        }
    }

    /* initializing stuff... */
    quit = FALSE;
    scene_time = 0;
    input = input_joystick_available() ? input_create_joystick() : input_create_keyboard(NULL);

    title = font_create(4);
    font_set_text(title, lang_get("CREDITS_TITLE"));
    title->position.x = (VIDEO_SCREEN_W - strlen(font_get_text(title))*font_get_charsize(title).x)/2;
    title->position.y = 5;

    back = font_create(8);
    font_set_text(back, lang_get("CREDITS_KEY"));
    back->position.x = 10;
    back->position.y = VIDEO_SCREEN_H - font_get_charsize(back).y - 5;

    text = font_create(8);
    font_set_text(text,
        "<color=ffff00>%s</color>\n\n"
        "%s\n\n"
        "\n"
        "<color=ffff00>%s</color>\n\n"
        "%s\n\n"
        "\n"
        "<color=ffff00>%s</color>\n\n"
        "%s\n\n"
        "%s\n\n"
        "%s\n\n"
        "\n",
        line[0], line[1], line[2], line[3], line[4], line[5], line[6], line[7]
    );
    text->position.x = 10;
    text->position.y = VIDEO_SCREEN_H;
    for(line_count=0,p=font_get_text(text); *p; p++)
        line_count += (*p == '\n') ? 1 : 0;

    box = image_create(VIDEO_SCREEN_W, 30);
    image_clear(box, image_rgb(0,0,0));

    bg = image_load(CREDITS_BGFILE);
    bgpos = v2d_new(0,0);

    fadefx_in(image_rgb(0,0,0), 1.0);
}


/*
 * credits_release()
 * Releases the scene
 */
void credits_release()
{
    image_unref(CREDITS_BGFILE);
    image_destroy(box);

    font_destroy(title);
    font_destroy(text);
    font_destroy(back);

    input_destroy(input);
}


/*
 * credits_update()
 * Updates the scene
 */
void credits_update()
{
    double dt = game_delta();
    scene_time += dt;

    /* background movement */
    bgpos.x -= (VIDEO_SCREEN_W/2)*dt;
    bgpos.y -= (VIDEO_SCREEN_H/2)*dt;

    /* text movement */
    text->position.y -= (3*font_get_charsize(text).y) * dt;
    if(text->position.y < -line_count*font_get_charsize(text).y)
        text->position.y = VIDEO_SCREEN_H;

    /* quit */
    if(!quit && scene_time >= 0.75 && input_button_pressed(input, IB_FIRE3)) {
        sound_play( sound_load("samples/select.wav") );
        quit = TRUE;
    }
    if(quit) {
        if(fadefx_over()) {
            scenestack_pop();
            scenestack_push(storyboard[SCENE_MENU]);
            return;
        }
        fadefx_out(image_rgb(0,0,0), 1.0);
    }
}



/*
 * credits_render()
 * Renders the scene
 */
void credits_render()
{
    int i, j;
    int bx = (((int)bgpos.x % bg->w) + bg->w) % bg->w;
    int by = (((int)bgpos.y % bg->h) + bg->h) % bg->h;
    v2d_t cam = v2d_new(VIDEO_SCREEN_W/2, VIDEO_SCREEN_H/2);

    for(i=-1; i<=VIDEO_SCREEN_W/bg->w; i++) {
        for(j=-1; j<=VIDEO_SCREEN_H/bg->h; j++)
            image_blit(bg, video_buffer, 0, 0, bx+bg->w*i, by+bg->h*j, bg->w, bg->h);
    }

    font_render(text, cam);
    image_blit(box, video_buffer, 0, 0, 0, 0, box->w, box->h);
    image_blit(box, video_buffer, 0, 0, 0, VIDEO_SCREEN_H-20, box->w, box->h);
    font_render(title, cam);
    font_render(back, cam);
}




/* private functions */
char *format_dev(char *str) /* explodes a "position1;position2;...;positionN" string */
{
    static char buf[1024];
    char *p;

    p = strtok(str, ";");
    sprintf(buf, "- %s", p);
    while(NULL != (p = strtok(NULL, ";")))
        sprintf(buf, "%s\n- %s", buf, p);

    return buf;
}
