/*
 * lang.c - language/translation module
 * Copyright (C) 2009  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "util.h"
#include "lang.h"



/*
 * lang_init()
 * Initializes the language module
 */
void lang_init()
{
    logfile_message("Initializing the language module");
    lang_loadfile(DEFAULT_LANGUAGE_FILEPATH);
    logfile_message("lang_init() ok!");
}


/*
 * lang_release()
 * Releases the language module
 */
void lang_release()
{
    logfile_message("Releasing the language module");
    logfile_message("lang_release() ok!");
}


/*
 * lang_loadfile()
 * Loads a language definition file
 */
void lang_loadfile(char *filepath)
{
    FILE *fp;
    char abs_path[1024];
    char line[1024], *p;
    char key[64], tmp[1024], data[1024], *s;
    int n;
    int ver, subver, wipver;

    lang_readcompatibility(filepath, &ver, &subver, &wipver);
    if(!(GAME_VERSION == ver && GAME_SUB_VERSION == subver && GAME_WIP_VERSION == wipver))
        fatal_error("FATAL ERROR: \"%s\" (c=%d.%d.%d) is NOT compatible with Open Sonic %d.%d.%d", filepath, ver, subver, wipver, GAME_VERSION, GAME_SUB_VERSION, GAME_WIP_VERSION);

    logfile_message("lang_loadfile(\"%s\")...", filepath);
    resource_filepath(abs_path, filepath, sizeof(abs_path), RESFP_READ);
    if(NULL != (fp=fopen(abs_path, "r"))) {
        while(fgets(line, sizeof(line), fp)) {

            for(p=line; *p && isspace((unsigned char)*p); p++);
            if(!*p) continue;
            sscanf(p, "%63s%n", key, &n);

            if(key[0] != '#') {
                for(p+=n; *p && isspace((unsigned char)*p); p++);
                str_cpy(tmp, p, sizeof(tmp));
                str_trim(data, tmp);

                s = (char*)hash_search(key, DATATYPE_STRING);
                if(s == NULL) {
                    s = mallocx(sizeof(data)); /* we must not use str_dup() here */
                    strcpy(s, data);
                    hash_insert(key, (void*)s, DATATYPE_STRING);
                }
                else
                    strcpy(s, data);
            }

        }
        fclose(fp);
        logfile_message("lang_loadfile(\"%s\") ok!", filepath);
    }
    else
        fatal_error("FATAL ERROR: couldn't load language file \"%s\"", filepath);
}


/*
 * lang_readstring()
 * Reads the contents of the desired key directly from the
 * language file (without loading it into memory)
 */
void lang_readstring(char *filepath, char *desired_key, char *str, int str_size)
{
    FILE *fp;
    char abs_path[1024];
    char line[1024], *p;
    char key[64], tmp[1024], data[1024] = "null";
    int n;

    resource_filepath(abs_path, filepath, sizeof(abs_path), RESFP_READ);
    if(NULL != (fp=fopen(abs_path, "r"))) {
        while(fgets(line, sizeof(line), fp)) {

            for(p=line; *p && isspace((unsigned char)*p); p++);
            if(!*p) continue;
            sscanf(p, "%63s%n", key, &n);

            if(strcmp(key, desired_key) == 0) {
                for(p=line+n; *p && isspace((unsigned char)*p); p++);
                str_cpy(tmp, p, sizeof(tmp));
                str_trim(data, tmp);
                break;
            }

        }
        str_cpy(str, data, str_size);
        fclose(fp);
    }
    else
        fatal_error("FATAL ERROR: lang_readstring(\"%s\", \"%s\") failed", filepath, desired_key);
}


/*
 * lang_getstring()
 * Retrieves some string from the language definition file
 */
void lang_getstring(char *desired_key, char *str, int str_size)
{
    char *p = (char*)hash_search(desired_key, DATATYPE_STRING);
    str_cpy(str, p ? p : "null", str_size);
}



/*
 * lang_get()
 * Like lang_getstring(), but returns the string as a static char*
 */
char *lang_get(char *desired_key)
{
    static char buf[1024];
    lang_getstring(desired_key, buf, sizeof(buf));
    return buf;
}



/*
 * lang_readcompatibility()
 * Language files are made for specific game versions
 */
void lang_readcompatibility(char *filename, int *ver, int *subver, int *wipver)
{
    char compat[32];
    lang_readstring(filename, "LANG_COMPATIBILITY", compat, sizeof(compat));
    if(sscanf(compat, "%d.%d.%d", ver, subver, wipver) < 3)
        *ver = *subver = *wipver = 0;
}
