/*
 * langselect.c - language selection screen
 * Copyright (C) 2009  Alexandre Martins < alemartf [.a.t.] gmail [.d.o.t.] com >
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <math.h>
#include "langselect.h"
#include "scene.h"
#include "util.h"
#include "video.h"
#include "audio.h"
#include "font.h"
#include "actor.h"
#include "lang.h"
#include "input.h"


/* private data */
#define LANG_BGFILE             "images/squarebg.png"
#define LANG_MAXPERPAGE         8
typedef struct {
    char title[128];
    char filepath[512];
} lngdata_t;

static image_t *bg;
static v2d_t bgpos;
static int quit;
static int lngcount;
static font_t *title[2], **lngfnt[2];
static lngdata_t *lngdata;
static int option; /* current option: 0..n-1 */
static actor_t *icon, *sonic;
static input_t *input;
static double scene_time;

/* private functions */
static void load_lang_list();
static void release_lang_list();
static int dirfill(const char *filename, int attrib, void *param);
static int dircount(const char *filename, int attrib, void *param);
static int sort_cmp(const void *a, const void *b);






/* public functions */

/*
 * langselect_init()
 * Initializes the scene
 */
void langselect_init()
{
    option = 0;
    quit = FALSE;
    scene_time = 0;
    input = input_joystick_available() ? input_create_joystick() : input_create_keyboard(NULL);

    title[0] = font_create(4);
    title[1] = font_create(4);
    font_set_text(title[0], "SELECT YOUR");
    font_set_text(title[1], "LANGUAGE");
    title[0]->position.x = (VIDEO_SCREEN_W - strlen(font_get_text(title[0]))*font_get_charsize(title[0]).x)/2;
    title[0]->position.y = 10;
    title[1]->position.x = (VIDEO_SCREEN_W - strlen(font_get_text(title[1]))*font_get_charsize(title[1]).x)/2;
    title[1]->position.y = title[0]->position.y + font_get_charsize(title[1]).y + 10;

    bg = image_load(LANG_BGFILE);
    bgpos = v2d_new(0,0);

    icon = actor_create();
    actor_change_animation(icon, sprite_get_animation("SD_TITLEFOOT", 0));
    sonic = actor_create();
    actor_change_animation(sonic, sprite_get_animation("SD_SONIC", 24));
    sonic->position = v2d_new(VIDEO_SCREEN_W-25, VIDEO_SCREEN_H-10);
    sonic->mirror = IF_HFLIP;

    load_lang_list();
    fadefx_in(image_rgb(0,0,0), 1.0);
}


/*
 * langselect_release()
 * Releases the scene
 */
void langselect_release()
{
    release_lang_list();
    image_unref(LANG_BGFILE);

    actor_destroy(sonic);
    actor_destroy(icon);
    font_destroy(title[0]);
    font_destroy(title[1]);
    input_destroy(input);
}


/*
 * langselect_update()
 * Updates the scene
 */
void langselect_update()
{
    double dt = game_delta();
    scene_time += dt;

    /* background movement */
    bgpos.x -= (VIDEO_SCREEN_W/2)*dt;
    bgpos.y -= (VIDEO_SCREEN_H/2)*dt;

    /* menu option */
    icon->position = lngfnt[0][option]->position;
    icon->position.x -= 15;
    icon->position.x += 5*cos(2*PI * scene_time);
    if(!quit && scene_time >= 0.75) {
        if(input_button_pressed(input, IB_DOWN)) {
            option = (option+1)%lngcount;
            sound_play( sound_load("samples/choose.wav") );
        }
        if(input_button_pressed(input, IB_UP)) {
            option = (((option-1)%lngcount)+lngcount)%lngcount;
            sound_play( sound_load("samples/choose.wav") );
        }
        if(input_button_pressed(input, IB_FIRE1) || input_button_pressed(input, IB_FIRE3)) {
            logfile_message("Loading language \"%s\", \"%s\"", lngdata[option].title, lngdata[option].filepath);
            lang_loadfile(DEFAULT_LANGUAGE_FILEPATH); /* just in case of missing strings... */
            lang_loadfile(lngdata[option].filepath);
            sound_play( sound_load("samples/select.wav") );
            quit = TRUE;
        }
        if(input_button_pressed(input, IB_FIRE4)) {
            sound_play( sound_load("samples/return.wav") );
            quit = TRUE;
        }
    }

    /* quit */
    if(quit) {
        if(fadefx_over()) {
            scenestack_pop();
            scenestack_push(storyboard[SCENE_MENU]);
            return;
        }
        fadefx_out(image_rgb(0,0,0), 1.0);
    }
}



/*
 * langselect_render()
 * Renders the scene
 */
void langselect_render()
{
    int i, j;
    int bx = (((int)bgpos.x % bg->w) + bg->w) % bg->w;
    int by = (((int)bgpos.y % bg->h) + bg->h) % bg->h;
    v2d_t cam = v2d_new(VIDEO_SCREEN_W/2, VIDEO_SCREEN_H/2);

    for(i=-1; i<=VIDEO_SCREEN_W/bg->w; i++) {
        for(j=-1; j<=VIDEO_SCREEN_H/bg->h; j++)
            image_blit(bg, video_buffer, 0, 0, bx+bg->w*i, by+bg->h*j, bg->w, bg->h);
    }

    font_render(title[0], cam);
    font_render(title[1], cam);

    for(i=0; i<lngcount; i++) {
        if(i/LANG_MAXPERPAGE == option/LANG_MAXPERPAGE)
            font_render(lngfnt[option==i ? 1 : 0][i], cam);
    }

    actor_render(icon, cam);
    actor_render(sonic, cam);
}






/* private methods */

/* reads the quest list from the quest/ folder */
void load_lang_list()
{
    int i, j, deny_flags = FA_DIREC | FA_LABEL, c = 0;
    int max_paths;
    char path[] = "languages/*.lng";
    char abs_path[2][1024];

    logfile_message("load_lang_list()");

    /* official and $HOME quests */
    absolute_filepath(abs_path[0], path, sizeof(abs_path[0]));
    home_filepath(abs_path[1], path, sizeof(abs_path[1]));
    max_paths = (strcmp(abs_path[0], abs_path[1]) == 0) ? 1 : 2;

    /* loading language data */
    lngcount = 0;
    for(j=0; j<max_paths; j++)
        for_each_file_ex(abs_path[j], 0, deny_flags, dircount, NULL);

    lngdata = mallocx(lngcount * sizeof(lngdata_t));
    for(j=0; j<max_paths; j++)
        for_each_file_ex(abs_path[j], 0, deny_flags, dirfill, (void*)&c);
    qsort(lngdata, lngcount, sizeof(lngdata_t), sort_cmp);

    /* fatal error */
    if(lngcount == 0)
        fatal_error("FATAL ERROR: no language files were found! Please reinstall the game.");
    else
        logfile_message("%d languages found.", lngcount);

    /* other stuff */
    lngfnt[0] = mallocx(lngcount * sizeof(font_t*));
    lngfnt[1] = mallocx(lngcount * sizeof(font_t*));
    for(i=0; i<lngcount; i++) {
        lngfnt[0][i] = font_create(8);
        lngfnt[1][i] = font_create(8);
        font_set_text(lngfnt[0][i], "%2d. %s", i+1, lngdata[i].title);
        font_set_text(lngfnt[1][i], "<color=ffff00>%2d. %s</color>", i+1, lngdata[i].title);
        lngfnt[0][i]->position = v2d_new(25, 80 + 20*(i%LANG_MAXPERPAGE));
        lngfnt[1][i]->position = v2d_new(25, 80 + 20*(i%LANG_MAXPERPAGE));
    }
}

int dirfill(const char *filename, int attrib, void *param)
{
    int *c = (int*)param;
    int ver, subver, wipver;

    lang_readcompatibility((char*)filename, &ver, &subver, &wipver);
    if(ver == GAME_VERSION && subver == GAME_SUB_VERSION && wipver == GAME_WIP_VERSION) {
        str_cpy(lngdata[*c].filepath, (char*)filename, sizeof( lngdata[*c].filepath ));
        lang_readstring((char*)filename, "LANG_LANGUAGE", lngdata[*c].title, sizeof( lngdata[*c].title ));
        (*c)++;
    }
    else
        logfile_message("Warning: language file \"%s\" (compatibility: %d.%d.%d) isn't compatible with this version of the game (%d.%d.%d)", filename, ver, subver, wipver, GAME_VERSION, GAME_SUB_VERSION, GAME_WIP_VERSION);

    return 0;
}

int dircount(const char *filename, int attrib, void *param)
{
    int ver, subver, wipver;
    lang_readcompatibility((char*)filename, &ver, &subver, &wipver);
    if(ver == GAME_VERSION && subver == GAME_SUB_VERSION && wipver == GAME_WIP_VERSION)
        lngcount++;

    return 0;
}

/* releases the language list */
void release_lang_list()
{
    int i;

    logfile_message("release_lang_list()");

    for(i=0; i<lngcount; i++) {
        font_destroy(lngfnt[0][i]);
        font_destroy(lngfnt[1][i]);
    }

    free(lngdata);
    lngcount = 0;
}


/* comparator */
int sort_cmp(const void *a, const void *b)
{
    lngdata_t *q[2] = { (lngdata_t*)a, (lngdata_t*)b };
    if(str_icmp(q[0]->title, "English") == 0) return -1;
    if(str_icmp(q[1]->title, "English") == 0) return 1;
    return str_icmp(q[0]->title, q[1]->title);
}

